// Protocol Buffers - Google's data interchange format
// Copyright 2008 Google Inc.  All rights reserved.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file or at
// https://developers.google.com/open-source/licenses/bsd

// Test that proto2 compiler can generate valid code when the enum value
// is INT_MAX. Note that this is a compile-only test and this proto is not
// referenced in any C++ code.
syntax = "proto2";

package protobuf_unittest;

message TestLargeEnumValue {
  enum EnumWithLargeValue {
    VALUE_1 = 1;
    VALUE_MAX = 0x7fffffff;
  }
}
