// Protocol Buffers - Google's data interchange format
// Copyright 2023 Google Inc.  All rights reserved.
//
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file or at
// https://developers.google.com/open-source/licenses/bsd

syntax = "proto2";

package protobuf_editions_test.golden;

import "google/protobuf/editions/golden/simple_proto2.proto";

message AnotherMessage {
  optional SimpleProto2 field = 1;
}
